import datetime
import os
import argparse
import subprocess
from hashlib import md5
from typing import Union

from bs4 import BeautifulSoup

from config import DANMAKU_EXEC, VIDEO_RESOLUTION, DANMAKU_SPEED, DANMAKU_FONT_NAME, DANMAKU_FONT_SIZE, \
    DANMAKU_USE_DANMU2ASS, DANMAKU_USE_DANMAKUFACTORY
from exception.danmaku import NoDanmakuException, DanmakuFormatErrorException
from util.file import check_file_exist


def get_file_start(file: Union[os.PathLike[str], str]) -> float:
    with open(file, "r", encoding="utf-8") as f:
        soup = BeautifulSoup("".join(f.readlines()), "lxml")
        danmaku_item = soup.find("d")
        if danmaku_item is None:
            # 没有弹幕？
            raise NoDanmakuException()
        danmaku_info = danmaku_item["p"]
        split_info = danmaku_info.split(",")
        if len(split_info) < 5:
            raise DanmakuFormatErrorException()
        bias_sec = float(split_info[0])
        bias_ts_ms = int(split_info[4])
        return bias_ts_ms / 1000 - bias_sec


def diff_danmaku_files(base_file: Union[os.PathLike[str], str], file: Union[os.PathLike[str], str]) -> float:
    return get_file_start(file) - get_file_start(base_file)


def danmaku_to_subtitle(file: Union[os.PathLike[str], str], time_shift: float):
    new_subtitle_name = md5(file.encode("utf-8")).hexdigest() + ".ass"
    if DANMAKU_USE_DANMAKUFACTORY:
        process = danmaku_to_subtitle_use_danmaku_factory(file, time_shift, new_subtitle_name)
    elif DANMAKU_USE_DANMU2ASS:
        process = danmaku_to_subtitle_use_danmu2ass(file, time_shift, new_subtitle_name)
    else:
        return
    process.wait()
    return new_subtitle_name


def danmaku_to_subtitle_use_danmaku_factory(file: Union[os.PathLike[str], str], time_shift: float, new_subtitle_name: str):
    return subprocess.Popen((
        DANMAKU_EXEC, "--ignore-warnings",
        "-r", str(VIDEO_RESOLUTION), "-s", str(DANMAKU_SPEED), "-f", "5",
        "-S", str(DANMAKU_FONT_SIZE), "-N", str(DANMAKU_FONT_NAME), "--showmsgbox", "FALSE",
        "-O", "255", "-L", "1", "-D", "0",
        "-o", "ass", new_subtitle_name, "-i", file, "-t", str(time_shift)
    ))


def danmaku_to_subtitle_use_danmu2ass(file: Union[os.PathLike[str], str], time_shift: float, new_subtitle_name: str):
    (_w, _h) = VIDEO_RESOLUTION.split("x")
    return subprocess.Popen((
        DANMAKU_EXEC, "--force", "-a", "1", "-d", str(DANMAKU_SPEED), "--font", str(DANMAKU_FONT_NAME),
        "--font-size", str(DANMAKU_FONT_SIZE), "--lane-size", str(DANMAKU_FONT_SIZE), "--width", _w, "--height", _h,
        "-o", new_subtitle_name, "-p", "1", "--time-offset", str(time_shift), "--width-ratio", "1", file
    ))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("base", help="以此为标准")
    parser.add_argument("file", nargs="+", help="需要对齐的文件")
    args = parser.parse_args()
    check_file_exist(args.base)
    base_start_ts = get_file_start(args.base)
    base_start = datetime.datetime.fromtimestamp(base_start_ts)
    print("[+] 基准文件[{:s}]，开始时间：{:.3f}，时间：{}".format(args.base, base_start_ts, base_start))
    for _file in args.file:
        check_file_exist(_file)
        file_start_ts = get_file_start(_file)
        diff_sec = file_start_ts - base_start_ts
        print("[+] 待调整文件[{:s}]，开始时间：{:.3f}，偏差：{:.3f}".format(_file, file_start_ts, diff_sec))
